@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: silcschedule.c,v 1.41.2.2 2002/12/29 21:27:38 priikone Exp $ */
+/* $Id: silcschedule.c,v 1.41.2.4 2003/01/16 17:45:14 priikone Exp $ */
 
 #include "silcincludes.h"
 #include "silcschedule_i.h"
@@ -399,12 +399,13 @@ static void silc_schedule_dispatch_nontimeout(SilcSchedule schedule)
 
       task = schedule->generic_queue->task;
       while(1) {
-	/* Validity of the task is checked always before and after
+	/* Validity of the task and fd is checked always before and after
 	   execution beacuse the task might have been unregistered
 	   in the callback function, ie. it is not valid anymore. */
 
 	/* Is the task ready for reading */				
-	if (task->valid && schedule->fd_list[i].revents & SILC_TASK_READ) {
+	if (task->valid && schedule->fd_list[i].revents & SILC_TASK_READ &&
+	    fd == schedule->fd_list[i].fd) {
 	  silc_mutex_unlock(schedule->generic_queue->lock);
 	  SILC_SCHEDULE_UNLOCK(schedule);
 	  task->callback(schedule, schedule->app_context,
@@ -414,7 +415,8 @@ static void silc_schedule_dispatch_nontimeout(SilcSchedule schedule)
 	}
 
 	/* Is the task ready for writing */				
-	if (task->valid && schedule->fd_list[i].revents & SILC_TASK_WRITE) {
+	if (task->valid && schedule->fd_list[i].revents & SILC_TASK_WRITE &&
+	    fd == schedule->fd_list[i].fd) {
 	  silc_mutex_unlock(schedule->generic_queue->lock);
 	  SILC_SCHEDULE_UNLOCK(schedule);
 	  task->callback(schedule, schedule->app_context,
