@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: server_util.c,v 1.65 2002/10/09 16:51:28 priikone Exp $ */
+/* $Id: server_util.c,v 1.72.2.3 2003/01/21 18:12:50 priikone Exp $ */
 
 #include "serverincludes.h"
 #include "server_internal.h"
@@ -42,7 +42,9 @@ silc_server_remove_clients_channels(SilcServer server,
   if (!client)
     return;
 
-  SILC_LOG_DEBUG(("Remove client from all channels"));
+  SILC_LOG_DEBUG(("Remove client %s from all channels",   
+		 client->nickname ? client->nickname :
+		  (unsigned char *)""));
 
   if (silc_hash_table_find(clients, client, NULL, NULL))
     silc_hash_table_del(clients, client);
@@ -923,7 +925,7 @@ bool silc_server_channel_has_local(SilcChannelEntry channel)
 
   silc_hash_table_list(channel->user_list, &htl);
   while (silc_hash_table_get(&htl, NULL, (void **)&chl)) {
-    if (!chl->client->router) {
+    if (SILC_IS_LOCAL(chl->client)) {
       silc_hash_table_list_reset(&htl);
       return TRUE;
     }
@@ -1456,8 +1458,8 @@ void silc_server_send_connect_notifys(SilcServer server,
   SILC_SERVER_SEND_NOTIFY(server, sock, SILC_NOTIFY_TYPE_NONE,
 			  ("Your connection is secured with %s cipher, "
 			   "key length %d bits",
-			   idata->send_key->cipher->name,
-			   idata->send_key->cipher->key_len));
+			   silc_cipher_get_name(idata->send_key),
+			   silc_cipher_get_key_len(idata->send_key)));
   SILC_SERVER_SEND_NOTIFY(server, sock, SILC_NOTIFY_TYPE_NONE,
 			  ("Your current nickname is %s",
 			   client->nickname));
@@ -1513,7 +1515,7 @@ void silc_server_kill_client(SilcServer server,
   /* Remove the client from all channels. This generates new keys to the
      channels as well. */
   silc_server_remove_from_channels(server, NULL, remote_client, FALSE, 
-				   NULL, TRUE);
+				   NULL, TRUE, TRUE);
 
   /* Remove the client entry, If it is locally connected then we will also
      disconnect the client here */
@@ -1714,3 +1716,245 @@ silc_server_find_socket_by_host(SilcServer server,
 
   return NULL;
 }
+
+/* This function can be used to match the invite and ban lists. */
+
+bool silc_server_inviteban_match(SilcServer server, SilcHashTable list,
+				 SilcUInt8 type, void *check)
+{
+  unsigned char *tmp = NULL;
+  SilcUInt32 len = 0, t;
+  SilcHashTableList htl;
+  SilcBuffer entry, idp = NULL;
+  bool ret = FALSE;
+
+  if (type < 1 || type > 3 || !check)
+    return FALSE;
+
+  if (type == 1) {
+    tmp = strdup((char *)check);
+    if (!tmp)
+      return FALSE;
+  }
+  if (type == 2) {
+    tmp = silc_pkcs_public_key_encode(check, &len);
+    if (!tmp)
+      return FALSE;
+  }
+  if (type == 3) {
+    idp = silc_id_payload_encode(check, SILC_ID_CLIENT);
+    if (!idp)
+      return FALSE;
+    tmp = idp->data;
+    len = idp->len;
+  }
+
+  /* Compare the list */
+  silc_hash_table_list(list, &htl);
+  while (silc_hash_table_get(&htl, (void **)&t, (void **)&entry)) {
+    if (type == t) {
+      if (type == 1) {
+	if (silc_string_match((char *)entry, tmp)) {
+	  ret = TRUE;
+	  break;
+	}
+      } else if (!memcmp(entry->data, tmp, len)) {
+	ret = TRUE;
+	break;
+      }
+    }
+  }
+  silc_hash_table_list_reset(&htl);
+
+  if (!idp)
+    silc_free(tmp);
+  silc_buffer_free(idp);
+  return ret;
+}
+
+static void silc_server_inviteban_dummy_dest(void *key, void *context,
+					     void *user_context)
+{
+  /* Nothing */
+}
+
+/* Process invite or ban information */
+
+void silc_server_inviteban_process(SilcServer server, SilcHashTable list,
+				   SilcUInt8 action, SilcArgumentPayload args)
+{
+  unsigned char *tmp;
+  SilcUInt32 type, len;
+  SilcBuffer tmp2;
+  SilcHashTableList htl;
+
+  SILC_LOG_DEBUG(("Processing invite/ban for %s action",
+		  action == 0x00 ? "ADD" : "DEL"));
+
+  /* Add the information to invite list */
+  if (action == 0x00) {
+    /* Traverse all arguments and add to the hash table according to
+       their type. */
+    tmp = silc_argument_get_first_arg(args, &type, &len);
+    while (tmp) {
+      if (type == 1) {
+	/* Invite string.  Get the old invite string from hash table
+	   and append this at the end of the existing one. */
+	char *string = NULL;
+	silc_hash_table_find(list, (void *)1,
+			     NULL, (void **)&string);
+	silc_hash_table_del_ext(list, (void *)1, NULL, NULL, NULL, NULL,
+				silc_server_inviteban_dummy_dest, NULL);
+	if (!string)
+	  string = silc_calloc(len + 2, sizeof(*string));
+	else
+	  string = silc_realloc(string, sizeof(*string) *
+				(strlen(string) + len + 2));
+	memset(string + strlen(string), 0, len + 2);
+	if (tmp[len - 1] == ',')
+	  tmp[len - 1] = '\0';
+	strncat(string, tmp, len);
+	strncat(string, ",", 1);
+
+	/* Add new invite string to invite list */
+	silc_hash_table_add(list, (void *)1, string);
+
+      } else if (type == 2) {
+	/* Public key.  Check first if the public key is already on the
+	   list and ignore it if it is, otherwise, add it to hash table. */
+
+	/* Check if the public key is in the list already */
+	silc_hash_table_list(list, &htl);
+	while (silc_hash_table_get(&htl, (void **)&type, (void **)&tmp2)) {
+	  if (type == 2 && !memcmp(tmp2->data, tmp, len)) {
+	    tmp = NULL;
+	    break;
+	  }
+	}
+	silc_hash_table_list_reset(&htl);
+
+	/* Add new public key to invite list */
+	if (tmp) {
+	  tmp2 = silc_buffer_alloc_size(len);
+	  silc_buffer_put(tmp2, tmp, len);
+	  silc_hash_table_add(list, (void *)2, tmp2);
+	}
+
+      } else if (type == 3) {
+	/* Client ID */
+
+	/* Check if the ID is in the list already */
+	silc_hash_table_list(list, &htl);
+	while (silc_hash_table_get(&htl, (void **)&type, (void **)&tmp2)) {
+	  if (type == 3 && !memcmp(tmp2->data, tmp, len)) {
+	    tmp = NULL;
+	    break;
+	  }
+	}
+	silc_hash_table_list_reset(&htl);
+
+	/* Add new Client ID to invite list */
+	if (tmp) {
+	  tmp2 = silc_buffer_alloc_size(len);
+	  silc_buffer_put(tmp2, tmp, len);
+	  silc_hash_table_add(list, (void *)3, tmp2);
+	}
+      }
+
+      tmp = silc_argument_get_next_arg(args, &type, &len);
+    }
+  }
+
+  /* Delete information to invite list */
+  if (action == 0x01 && list) {
+    /* Now delete the arguments from invite list */
+    tmp = silc_argument_get_first_arg(args, &type, &len);
+    while (tmp) {
+      if (type == 1) {
+	/* Invite string.  Get the old string from hash table and delete
+	   the requested string. */
+	char *string = NULL, *start, *end, *n;
+
+	if (silc_hash_table_find(list, (void *)1, NULL, (void **)&string)) {
+	  if (!strncmp(string, tmp, strlen(string) - 1)) {
+	    silc_hash_table_del(list, (void *)1);
+	    string = NULL;
+	  } else {
+	    start = strstr(string, tmp);
+	    if (start && strlen(start) >= len) {
+	      end = start + len;
+	      n = silc_calloc(strlen(string) - len, sizeof(*n));
+	      strncat(n, string, start - string);
+	      strncat(n, end + 1, ((string + strlen(string)) - end) - 1);
+	      silc_hash_table_del(list, (void *)1);
+	      string = n;
+	    }
+	  }
+
+	  /* Add new invite string to invite list */
+	  if (string)
+	    silc_hash_table_add(list, (void *)1, string);
+	}
+
+      } else if (type == 2) {
+	/* Public key. */
+
+	/* Delete from the invite list */
+	silc_hash_table_list(list, &htl);
+	while (silc_hash_table_get(&htl, (void **)&type, (void **)&tmp2)) {
+	  if (type == 2 && !memcmp(tmp2->data, tmp, len)) {
+	    silc_hash_table_del_by_context(list, (void *)2, tmp2);
+	    break;
+	  }
+	}
+	silc_hash_table_list_reset(&htl);
+
+      } else if (type == 3) {
+	/* Client ID */
+
+	/* Delete from the invite list */
+	silc_hash_table_list(list, &htl);
+	while (silc_hash_table_get(&htl, (void **)&type, (void **)&tmp2)) {
+	  if (type == 3 && !memcmp(tmp2->data, tmp, len)) {
+	    silc_hash_table_del_by_context(list, (void *)3, tmp2);
+	    break;
+	  }
+	}
+	silc_hash_table_list_reset(&htl);
+      }
+
+      tmp = silc_argument_get_next_arg(args, &type, &len);
+    }
+  }
+}
+
+/* Destructor for invite or ban list entrys */
+
+void silc_server_inviteban_destruct(void *key, void *context,
+				    void *user_context)
+{
+  switch ((SilcUInt32)key) {
+  case 1:
+    /* Invite string */
+    silc_free(context);
+    break;
+  case 2:
+  case 3:
+    /* Public key/Channel ID SilcBuffer */
+    silc_buffer_free(context);
+    break;
+  default:
+    break;
+  }
+}
+
+/* Creates connections accoring to configuration. */
+
+void silc_server_create_connections(SilcServer server)
+{
+  silc_schedule_task_del_by_callback(server->schedule,
+				     silc_server_connect_to_router);
+  silc_schedule_task_add(server->schedule, 0,
+			 silc_server_connect_to_router, server, 0, 1,
+			 SILC_TASK_TIMEOUT, SILC_TASK_PRI_NORMAL);
+}
